\ Forth code to control four si4703 FM receiver devices.
\ The devices are connected to pins (0..15) on the Propeller
\ Device 0 pins are; 0-SDA 1-SCLK 2-Reset 3-GPIO2.
\ Device 1 pins are; 4-SDA 5-SCLK 6-Reset 7-GPIO2.
\ Device 2 pins are; 8-SDA 9-SCLK 10-Reset 11-GPIO2.
\ Device 3 pins are; 12-SDA 13-SCLK 14-Reset 15-GPIO2.
\ 
#4 CONSTANT si_DEVICES
\ variable to hold propeller cog driver runs on
0 VALUE si_cog

\ command buffer used by driver
QVARIABLE si_cmd
0. si_cmd 2!
  
\ Buffer to hold shadow copy of si_DEVICES device registers.
\ Each device has 16 registers each 16 bits in size.
\ Device 0 uses first 32 bytes, device 1 uses next
\ 32 bytes...
si_DEVICES #32 * QBYTES siregbuf

\ device state constants
0 CONSTANT si_NOT_FOUND
1 CONSTANT si_FOUND
#2 CONSTANT si_TUNE
#3 CONSTANT si_SEEK
 
\ holds state for each device
VARIABLE si_devstates si_DEVICES 2* ALLOT
si_devstates si_DEVICES 2* ERASE

\ save address of selected device register buffer
siregbuf VALUE sireg

\ used to select correct device register buffer
: si-setbuffer ( uDeviceIdx -- )
  DUP si_DEVICES U< IF
    #32 * siregbuf + 
    TO sireg
  ELSE
    DROP
  THEN ;
  
: si-stop ( -- )
  si_cog IF
    si_cog COG-STOP
    0 TO si_cog
  THEN ;
  
: si-reg! ( uValue, uRegIdx -- )  
  2* sireg + ! ;   
    
: si-reg@ ( uRegIdx -- uValue )  
  2* sireg + @ ;   
    
: si-read ( adrReg uCount -- fWasRead )
  si_cmd #2 COG-EX ;
  
: si-write ( adrReg uCount -- fWasWritten )
  si_cmd #3 COG-EX ;
  
: si-device@ ( -- uDeviceIdx )
  si_cmd #5 COG-EX ;
  
: si-device! ( uDeviceIdx -- )
  DUP si_cmd #4 COG-EX 
  si-setbuffer ;

\ reset selected device
: si-reset ( -- )
  si_cog IF
    \ reset the si4703 and init i2c buss
    si_cmd 1 COG-EX 
  THEN ;

: si-setdevstate ( uNewState -- )
  si_devstates si-device@ 2* + ! ;
  
: si-getdevstate ( -- uNewState )
  si_devstates si-device@ 2* + @  ;  
  
\ read all the register for the selected radio device  
: si-readRegisters ( -- fWasRead )
  si-getdevstate si_NOT_FOUND <> IF 
    sireg #15 si-read
  ELSE
    FALSE
  THEN ;    

: si-init ( -- )
  \ if cog is running try to reset device
  si_cog IF
    \ reset the si4703 and init i2c buss
    si-reset 
    \ read control register to shadow registers
    si-readRegisters DUP IF
      DROP
      \ set register 7 to enable crytal osc
      $8100 #7 si-reg!
      \ update device
      sireg #6 si-write DUP IF
        DROP 
        #500 MS
        \ set register 2 to enable device
        $0001 #2 si-reg!
        sireg #0 si-write
        #100 MS
      THEN
    THEN
  ELSE
    \ cog is not running
    FALSE  
  THEN ;
  
\ reset all radio devices
: si-init-all ( -- uDeviceCount )
  \ reset all radio devices
  si_cog IF
    si_DEVICES 0 DO
      I si-device! si-reset
      \ set device is found so read registers will not fail
      si_FOUND si-setdevstate 
    LOOP
    \ init shadow registers for each device
    #5 MS
    si_DEVICES 0 DO
      I si-device! 
      si-readRegisters 0= IF
        \ failed to read registers, indicate device is not available
        si_NOT_FOUND si-setdevstate
      THEN
    LOOP
    \ enable crystal osc for each device
    si_DEVICES 0 DO
      I si-device!
      \ if device registers where read, initialize osc
      si-getdevstate si_FOUND = IF 
        $8100 #7 si-reg!
        \ update device
        sireg #6 si-write NOT IF
          \ falied to initialize device
          si_NOT_FOUND si-setdevstate    
        THEN
      THEN  
    LOOP
    \ delay for about a half a second to allow osc to become stable
    #500 MS
    \ last enable device, and count active radios
    0
    si_DEVICES 0 DO
      I si-device!
      \ if device registers where read, enable device
      si-getdevstate si_FOUND = IF 
        $0001 #2 si-reg!
        sireg #0 si-write IF
          1+
        ELSE
          \ falied to initialize device
          si_NOT_FOUND si-setdevstate    
        THEN
      THEN  
    LOOP
    #100 MS
  ELSE
    0
  THEN ;    

\ start the si4703 driver COG, if it is running stop it then restart
: si-run ( -- fFoundRadio)
  si-stop
  si_cmd #30 COG-LOAD 
  TO si_cog si-init-all ;

\ set audio mute on/off, setting bit high disables mute
: si-mute ( fMuteAudio -- fWasSet )
  si-getdevstate si_NOT_FOUND <> IF
    #2 si-reg@ SWAP 
    IF
      $BFFF AND
    ELSE
      $4000 OR
    THEN
    #2 si-reg! 
    sireg 0 si-write
  ELSE
    FALSE
  THEN ;    
  
: si-mute? ( -- fIsMutted )
  #2 si-reg@
  $4000 AND 0= ;
  
\ set force mono audio on/off, set bit high to force mono  
: si-mono ( fMono -- fWasSet )
  si-getdevstate si_NOT_FOUND <> IF
    #2 si-reg@ SWAP
    IF
      $2000 OR
    ELSE
      $DFFF AND
    THEN
    #2 si-reg! 
    sireg 0 si-write  
  ELSE
    FALSE
  THEN ;    

: si-mono? ( -- fIsMono )
  #2 si-reg@
  $2000 AND 0> ;
   
: si-vol! ( uVol -- fWasSet )
  si-getdevstate si_NOT_FOUND <> IF
    FALSE si-mute IF
      ( uVol )
      $0F AND
      #5 si-reg@
      $FFF0 AND
      ( uVol' uReg5' )
      OR
      #5 si-reg!
      sireg #3 si-write
    ELSE
      FALSE
    THEN    
  ELSE
    FALSE
  THEN ;    
  
: si-vol@ ( -- uVol )
  #5 si-reg@
  $0F AND ;


  
  \ update status register (0Ah) and return value 
: si-status ( -- uValue | -1 )
  si-getdevstate si_NOT_FOUND <> IF
    sireg 0 si-read DROP
    $0A si-reg@
  ELSE
    -1  
  THEN ;  

\ tune radio to specified station  
: si-tune ( uFreq -- fWasSet )
  si-getdevstate si_NOT_FOUND <> IF
    \ if we set a prior tune we need to cancel it
    #3 si-reg@ $8000 AND IF
      ." clear tune bit" CR
       \ we have a prior tune, clear tune bit  
       #3 si-reg@ $7FFF AND
       #3 si-reg!
       sireg 1 si-write
       1 MS
    ELSE
      TRUE   
    THEN
    ( uFreq fOK )
    OVER
    ( uFreq fOK uFreq )
    #875 #1080 WITHIN AND
    ( uFreq fOK' )
    IF
      \ calculate the channel index
      #875 - 2/
      $8000 +
      \ write fm channel index to device register
      #3 si-reg!
      sireg 1 si-write IF
        \ wait for the channel to tune
        BEGIN
          #1 MS
          si-status $4000 AND
        UNTIL
        \ we completed tune so clear tune bit
        #3 si-reg@ $7FFF AND
        #3 si-reg!
        sireg 1 si-write
      ELSE
        FALSE
      THEN  
    ELSE
      DROP
      FALSE
    THEN  
  ELSE
    DROP
    FALSE
  THEN ;  
      
: si-status ( -- uValue )
  sireg 0 si-read DROP
  $0A si-reg@ ;

\ seek to the next or prior rf channel
: (si-seek) ( fSeekUp -- fWasOk )
  si-getdevstate si_NOT_FOUND <> IF
    #2 si-reg@
    ( fSeekUp uRegVal )
    SWAP IF
      $0300
    ELSE
      $FDFF AND
      $0100
    THEN
    ( uRegVal uBitMsk )   
    OR #2 si-reg!
    sireg 0 si-write IF
      \ wait for the channel to seek, S/T bit high
      BEGIN
        #1 MS
        si-status $4000 AND
      UNTIL
      \ we completed tune so clear seek bit
      #2 si-reg@ $FEFF AND
      #2 si-reg!
      sireg 1 si-write
    ELSE
      FALSE  
    THEN  
  ELSE
    FALSE
  THEN ;    

  \ seek to the next rf channel
: si-seekup ( -- fWasOk )
  TRUE (si-seek) ;

\ seek to the prior rf channel
: si-seekdown ( -- fWasOk )
  FALSE (si-seek) ;
  
  
: si-channel ( -- uChannelIdx | -1 )
  si-getdevstate si_NOT_FOUND <> IF
    sireg 1 si-read DROP
    $0B si-reg@
  ELSE
    -1
  THEN ;
  
    
: si-seekth! ( uSeekThresholdDB -- fWasSet )
  si-getdevstate si_NOT_FOUND <> IF
    FALSE si-mute IF
      ( uSeekTh )
      $FF AND #8 LSHIFT
      #5 si-reg@
      $00FF AND
      ( uSeekTh' uReg5' )
      OR
      #5 si-reg!
      sireg #3 si-write
    ELSE
      FALSE
    THEN    
  ELSE
    FALSE
  THEN ;    
  
: si-seekth@ ( -- uVol )
  #5 si-reg@
  #8 RSHIFT ;


  
( **** TEST CODE TO EXERSIZE DEVICES RDS **** )


: si-rds. ( -- )
  \ wait for a valid rds group
  si-getdevstate si_NOT_FOUND <> IF
    BEGIN
      KEY? NOT
      si-status $8000 AND 0=
      AND
    WHILE  
      10 MS
    REPEAT
    KEY? NOT IF
      sireg #5 si-read DROP
      HEX
      $10 $0C DO
        I si-reg@
        S>D <# # # # # #> TYPE SPACE
      LOOP
      $11 si-reg@ $4000 AND IF
        CR ." time block" CR
      THEN
      BEGIN
        5 MS
        KEY?
        si-status $8000 AND 0=
        OR
      UNTIL
      DECIMAL
    THEN
  THEN ;  

( ******** TEST WORDS TO EXECSIZE RADIO DEVICES ******** )
        
: word.b  ( u - ) 
  BINARY 
  S>D <# #16 0 DO # LOOP #> TYPE SPACE 
  DECIMAL ;

: word.h  ( u - ) 
  HEX 
  S>D <# #4 0 DO # LOOP #> TYPE SPACE 
  DECIMAL ;

\ read all resisters from si4703 and send to console 
: si-dump ( -- )
  si-getdevstate si_NOT_FOUND <> IF
    si-readRegisters IF
      CR
      #16 0 DO
        I si-reg@ DUP
        I #2 U.R SPACE word.b word.h CR
      LOOP
    ELSE
      ." device was not found"
    THEN  
  THEN ;  

: si-status. ( -- )
  si-status word.b ;

: si-rssi. ( -- )
  si-status $FF AND . ." dBuV " ;


: si-channel. ( -- )
  si-channel 2* 875 +
  S>D <# # [CHAR] . HOLD #S #> 
  TYPE SPACE ." mhz " ;   
       

: si-test ( -- )
  \ enable rds
  si-getdevstate si_NOT_FOUND <> IF
    CR
    #4 si-reg@ $1000 OR
    #4 si-reg!
    sireg #3 si-write DROP
    BEGIN
      KEY? NOT
    WHILE
      si-rds. CR
    REPEAT 
  THEN ;           
  